#!/usr/bin/env python
# -*-coding:utf-8-*-

import os
from nltk import FreqDist, bigrams
import nltk
from math import log

from fenci.utils import read_training_content


def suggest_bmes(word):
    if len(word) == 1:
        return f'{word}/S'
    elif len(word) == 2:
        return f'{word[0]}/B {word[1]}/E'
    elif len(word) == 3:
        return f'{word[0]}/B {word[1]}/M {word[2]}/E'
    elif len(word) > 3:
        result = f'{word[0]}/B '
        for s in word[1:-1]:
            result += f'{s}/M '
        result += f'{word[-1]}/E'
        return result
    else:
        print(f'wrong word length !!!!')


def prepare_bmes_content(root, regexp):
    content = read_training_content(root, regexp)
    content_list = content.split()

    bmes_content_list = [suggest_bmes(w) for w in content_list]

    new_bmes_content_list = []

    for t in bmes_content_list:
        for s in t.split():
            new_bmes_content_list.append(nltk.tag.str2tuple(s))

    return new_bmes_content_list


def train_trans_matrix(root, regexp):
    """
    BB BM BE BS
    MB MM ME MS
    EB EM EE ES
    SB SM SE SS
    pBM = C(BM)/C(B)
    :return:
    """
    P_transMatrix = {'B': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                     'E': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                     'M': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                     'S': {'B': 0, 'E': 0, 'M': 0, 'S': 0}}
    bmes_content_list = prepare_bmes_content(root, regexp)

    bmes_list = [item[0][-1] + item[1][-1] for item in
                 bigrams(bmes_content_list)]

    result_fd = FreqDist(bmes_list)

    from itertools import product
    for key in map(lambda a: a[0] + a[1],
                   product(['B', 'M', 'E', 'S'], repeat=2)):
        if key in result_fd:
            a = key[0]
            b = key[1]
            P_transMatrix[a][b] = result_fd[key]

    new_P_transMatrix = {}

    for k in P_transMatrix:
        for k2 in P_transMatrix[k]:
            if P_transMatrix[k][k2] == 0:
                pass
            else:
                if k not in new_P_transMatrix:
                    new_P_transMatrix[k] = {}
                new_P_transMatrix[k][k2] = P_transMatrix[k][k2]

    return new_P_transMatrix


def train_trans_matrix_to_file(root, regexp, output_dir='.'):
    P_transMatrix = train_trans_matrix(root, regexp)

    for k, v in P_transMatrix.items():
        count = sum(v.values())
        for k2 in v:
            P_transMatrix[k][k2] = log(P_transMatrix[k][k2] / count)

    with open(os.path.join(output_dir, 'hmm/prob_trans.py'), 'wt',
              encoding='utf8') as f:
        print(f"""P={P_transMatrix}""", file=f)


def train_emit_matrix(root, regexp):
    P_emit = {'B': {}, 'E': {}, 'M': {}, 'S': {}}
    bmes_content_list = prepare_bmes_content(root, regexp)

    bmes_emit_list = [item[0][-1] + item[1][0] for item in
                      bigrams(bmes_content_list)]

    result_emit_fd = FreqDist(bmes_emit_list)

    for k, v in result_emit_fd.items():
        P_emit[k[0]][k[-1]] = v

    return P_emit


def train_emit_matrix_to_file(root, regexp, output_dir='.'):
    P_emit = train_emit_matrix(root, regexp)
    for k, v in P_emit.items():
        count = sum(v.values())
        for k2 in v:
            P_emit[k][k2] = log(P_emit[k][k2] / count)

    with open(os.path.join(output_dir, 'hmm/prob_emit.py'), 'wt',
              encoding='utf8') as f:
        print(f"""P={P_emit}""", file=f)


if __name__ == '__main__':
    root = 'icwb2-data/training'
    regexp = '(?!\.).*\.utf8'

    train_trans_matrix_to_file(root, regexp)
    train_emit_matrix_to_file(root, regexp)



#!/usr/bin/env python
# -*-coding:utf-8-*-


def test_load_dict():
    from fenci.segment import Segment
    s = Segment()

    res1 = s.tokenize(
        "机器学习是一门新型的计算机学科。")

    s.load_userdict('tests/test_dict.txt')

    res2 = s.tokenize("机器学习是一门新型的计算机学科。")

    assert '机器学习' not in res1
    assert '机器学习' in res2



#!/usr/bin/env python
# -*-coding:utf-8-*-

import re
import math
import threading
import logging
import os
import time

from nltk import FreqDist
from nltk.tokenize.api import TokenizerI

from .base import BaseSegment
from .hmm_segment import HMMSegment
from .utils import normalized_path, get_json_value, set_json_value
from . import __softname__
from .const import DEFAULT_DICT, DEFALUT_CACHE_NAME, CACHE_WRITING
from .utils import strdecode

logger = logging.getLogger(__name__)

re_userdict = re.compile('^(.+?)( [0-9]+)?( [a-z]+)?$')
re_eng = re.compile('[a-zA-Z0-9]')

re_han_default = re.compile(r"([\u4E00-\u9FD5a-zA-Z0-9+#&\._%\-]+)")
re_skip_default = re.compile(r"([\r\n|\s]+)")


class Segment(TokenizerI, BaseSegment):
    def __init__(self, dictionary=None, traning_root=None,
                 traning_regexp='.*\.txt'):
        self.lock = threading.RLock()

        self.training_root = traning_root
        self.training_regexp = traning_regexp

        if dictionary is None:
            self.dictionary = DEFAULT_DICT
            self.dictionary_type = 'default'
        else:
            self.dictionary = normalized_path(dictionary)
            self.dictionary_type = 'custom'

        self.word_fd = FreqDist()

        self.cache_file = DEFALUT_CACHE_NAME

        self.hmm_segment = HMMSegment(traning_root=traning_root,
                                      traning_regexp=traning_regexp,
                                      cache_file=self.cache_file)

        self.initialized = False
        self.tmp_dir = None

    def training(self, root=None, regexp=None):
        """
        根据已经分好词的内容来训练
        :param root:
        :param regexp:
        :return:
        """
        self.check_initialized()

        if root is None and self.training_root is None:
            raise Exception('please give the training data root')

        root = root if root is not None else self.training_root
        regexp = regexp if regexp is not None else self.training_regexp

        from fenci.utils import read_training_content
        content = read_training_content(root, regexp)

        content_list = content.split()

        fd = FreqDist(content_list)

        self.word_fd.update(fd)

    def training_hmm(self, root=None, regexp=None, update_dict=False):
        self.check_initialized()

        if root is None and self.training_root is None:
            raise Exception('please give the training data root')
        root = root if root is not None else self.training_root
        regexp = regexp if regexp is not None else self.training_regexp

        if update_dict:
            self.training(root, regexp)

        self.hmm_segment.training(root, regexp)

    def gen_word_fd(self, filename):
        word_fd = FreqDist()

        with open(filename, 'rt', encoding='utf8') as f:
            for line in f:
                word, freq = line.split()[:2]
                freq = int(freq)
                word_fd.update({word: freq})

        return word_fd

    def initialize(self):
        with self.lock:  # 字典在建造时的线程锁
            if self.initialized:  # 已经初始化了就不用初始化了
                return

            logger.debug("Building prefix dict from %s ..." % (
                    self.dictionary or 'the default dictionary'))
            t1 = time.time()

            cache_file = self._get_cache_file()

            # use cache data
            use_cache_data = False
            if os.path.isfile(cache_file):
                word_fd_timestamp = get_json_value(cache_file,
                                                   'word_fd_timestamp')
                if word_fd_timestamp:
                    if self.dictionary_type == 'custom':
                        if int(word_fd_timestamp) > os.path.getmtime(
                                self.dictionary):
                            use_cache_data = True
                    else:
                        use_cache_data = True

            if use_cache_data:
                logger.debug("Loading model from cache {0}".format(cache_file))

                word_fd = get_json_value(cache_file, 'word_fd')
                self.word_fd = FreqDist(word_fd)
            else:
                word_fd = self.gen_word_fd(self._get_dict_file())
                self.word_fd = FreqDist(word_fd)

                self.save_model(save_hmm=False)

            self.initialized = True
            logger.debug(
                "Loading model cost %.3f seconds." % (time.time() - t1))
            logger.debug("Prefix dict has been built succesfully.")

    def _get_dict_file(self):
        if self.dictionary == DEFAULT_DICT:
            from pkg_resources import resource_filename
            return resource_filename(__softname__, self.dictionary)
        else:
            return self.dictionary

    def get_DAG(self, sentence):
        self.check_initialized()

        DAG = {}
        N = len(sentence)
        for k in range(N):
            tmplist = []
            i = k
            frag = sentence[k]
            while i < N:
                if self.word_fd.get(frag, 0) > 0:
                    tmplist.append(i)
                i += 1
                frag = sentence[k:i + 1]
            if not tmplist:
                tmplist.append(k)
            DAG[k] = tmplist
        return DAG

    def calc(self, sentence, DAG, route):
        N = len(sentence)
        route[N] = (0, 0)

        logtotal = math.log(self.word_fd.N())
        for idx in range(N - 1, -1, -1):  # 逆序规划 选择一条整个路径频率最大的句子
            route[idx] = max(
                (math.log(self.word_fd.get(sentence[idx:x + 1]) or 1) -
                 logtotal + route[x + 1][0],
                 x) for x in DAG[idx])  # x 终点索引点 idx 考察开始点

    def __cut_DAG(self, sentence):
        DAG = self.get_DAG(sentence)
        route = {}
        self.calc(sentence, DAG, route)

        x = 0
        buf = ''
        N = len(sentence)
        while x < N:
            y = route[x][1] + 1
            l_word = sentence[x:y]
            if y - x == 1:
                buf += l_word  # 单字母或单字
            else:
                if buf:
                    if len(buf) == 1:  # 夹着的单字
                        yield buf
                        buf = ''
                    else:
                        if not self.word_fd.get(buf):  # 词典里找不到的词 用HMM来分
                            recognized = self.hmm_segment.cut(buf)
                            for t in recognized:
                                yield t
                        else:
                            for elem in buf:
                                yield elem
                        buf = ''

                yield l_word  # 找到的词优先输出
            x = y

        # 纯单字母或单字的情况
        if buf:
            if len(buf) == 1:
                yield buf
            elif not self.word_fd.get(buf):
                recognized = self.hmm_segment.cut(buf)
                for t in recognized:
                    yield t
            else:
                for elem in buf:
                    yield elem

    def tokenize(self, s):
        return self.lcut(s)

    def cut(self, sentence):
        """
        """
        sentence = strdecode(sentence)

        re_han = re_han_default
        re_skip = re_skip_default

        cut_block = self.__cut_DAG

        blocks = re_han.split(sentence)

        for blk in blocks:

            if not blk:  # 空白符号跳过
                continue
            if re_han.match(blk):  # 中文符号 核心分词在这里
                for word in cut_block(blk):
                    yield word
            else:
                tmp = re_skip.split(blk)
                for x in tmp:

                    if re_skip.match(x):  # 多个空白不分开
                        yield x
                    else:
                        for xx in x:  # 剩下来的全部分开
                            yield xx

    def lcut(self, sentence):
        return list(self.cut(sentence))

    def load_userdict(self, filename):
        self.check_initialized()

        with open(filename, 'rt', encoding='utf8') as f:
            for line in f:
                word, freq, tag = re_userdict.match(line).groups()
                word = word.strip()
                if freq is not None:
                    freq = freq.strip()
                else:
                    freq = 1

                self.add_word(word, freq)

    def add_word(self, word, freq=1):
        """
        Add a word to dictionary.
        freq and tag can be omitted, freq defaults to be a calculated value
        that ensures the word can be cut out.
        """
        self.check_initialized()
        word = strdecode(word)
        freq = int(freq)

        self.word_fd.update({word: freq})

    def save_model(self, save_hmm=False):
        wlock = CACHE_WRITING.get(self.dictionary, threading.RLock())
        CACHE_WRITING[self.dictionary] = wlock
        cache_file = self._get_cache_file()
        with wlock:
            logger.debug(
                "Dumping model to file cache {0}".format(cache_file))

            set_json_value(cache_file, 'word_fd', dict(self.word_fd))
            set_json_value(cache_file, 'word_fd_timestamp', int(time.time()))

            if save_hmm:
                self.hmm_segment.save_model()

        try:
            del CACHE_WRITING[self.dictionary]
        except KeyError:
            pass



#!/usr/bin/env python
# -*-coding:utf-8-*-

import os
import json
import logging
import re
from pathlib import Path

logger = logging.getLogger(__name__)


def normalized_path(path='.') -> str:
    """
    默认支持 ~ 符号

    返回的是字符串

    which default support the `~`
    """
    if isinstance(path, Path):
        return str(path.expanduser())
    elif isinstance(path, str):
        if path.startswith('~'):
            path = os.path.expanduser(path)
        return path
    else:
        raise TypeError


def strdecode(sentence):
    """
    字符串decode 只尝试utf8，和 gbk 否则将抛出异常
    :param sentence:
    :return:
    """
    if not isinstance(sentence, str):
        try:
            sentence = sentence.decode('utf-8')
        except UnicodeDecodeError:
            try:
                sentence = sentence.decode('gbk')
            except UnicodeDecodeError:
                logger.error('I have tried utf8 and gbk encoding, all failed.')
                raise UnicodeDecodeError
    return sentence


def write_json(file, data):
    """
    write data to json file
    """
    with open(file, 'w', encoding='utf8') as f:
        json.dump(data, f, indent=4, ensure_ascii=False)


def get_json_file(json_filename, default_data=None):
    """
    try get json file or auto-create the json file with default_data
    """
    default_data = default_data if default_data is not None else {}

    if not os.path.exists(json_filename):
        data = default_data
        write_json(json_filename, data)

    return json_filename


def get_json_data(json_filename, default_data=None):
    """
    get json file data
    """
    with open(get_json_file(json_filename, default_data=default_data),
              encoding='utf8') as f:
        data = json.load(f)
        return data


def get_json_value(json_filename, key, default_data=None):
    """
    get value by key in json file if your json file stored value as one dict.
    """
    data = get_json_data(json_filename, default_data=default_data)
    if not isinstance(data, dict):
        raise Exception(
            "the target json file must stored whole data as one dict.")

    return data.get(key)


def set_json_value(json_filename, key, value, default_data=None):
    """
    set value by key and value in json file if your json file stored value
    as one dict.
    """
    data = get_json_data(json_filename, default_data=default_data)

    if not isinstance(data, dict):
        raise Exception(
            "the target json file must stored whole data as one dict.")

    data[key] = value
    write_json(get_json_file(json_filename), data)


def find_trainning_files(root, regexp, **kwargs):
    items = []

    for dirname, subdirs, fileids in os.walk(root, **kwargs):
        for fileid in fileids:
            if re.match(regexp, fileid):
                items.append(os.path.join(dirname, *subdirs, fileid))
    return items


def read_training_content(root, regexp):
    return ''.join([open(file, encoding='utf8').read() for file in
                    find_trainning_files(root, regexp)])



#!/usr/bin/env python
# -*-coding:utf-8-*-

from abc import ABC, abstractmethod
import os
import tempfile


class BaseSegment(ABC):

    @abstractmethod
    def initialize(self):
        pass

    def check_initialized(self):
        if not self.initialized:
            self.initialize()

    def _get_cache_file(self):
        cache_file = os.path.join(self.tmp_dir or tempfile.gettempdir(),
                                  self.cache_file)
        self.tmp_dir = os.path.dirname(cache_file)
        return cache_file



#!/usr/bin/env python
# -*-coding:utf-8-*-

import logging
from fenci.segment import Segment

logging.basicConfig(level=logging.DEBUG)


def test_cut():
    segment = Segment()
    res = segment.lcut(
        '据 CNBC 报道，Google    前 CEO、Alphabet 前执行董事 Eric Schmidt 近日在参加旧金山的某高级私人活动时表示，未来十年将有两个截然不同的互联网：一个由美国领导，另一个由中国领导。。。')

    print(res)

    assert res == ['据', ' ', 'CNBC', ' ', '报道', '，', 'Google', '    ', '前', ' ',
                   'CEO', '、', 'Alphabet', ' ', '前', '执行', '董事', ' ', 'Eric',
                   ' ', 'Schmidt',
                   ' ', '近日', '在', '参加', '旧金山', '的', '某', '高级', '私人', '活动', '时',
                   '表示', '，', '未来', '十年', '将', '有', '两个', '截然不同', '的',
                   '互联网', '：', '一个', '由', '美国', '领导', '，', '另', '一个', '由', '中国',
                   '领导', '。', '。', '。']



#!/usr/bin/env python
# -*-coding:utf-8-*-

import os
from setuptools import setup, find_packages
import fenci

REQUIREMENTS = []
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='fenci',
    version=fenci.__version__,
    description='中文分词',
    url='https://github.com/a358003542/fenci',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='wanze',
    author_email='a358003542@gmail.com',
    maintainer='wanze',
    maintainer_email='a358003542@gmail.com',
    license='MIT',
    platforms='Linux,windows',
    keywords=['word_segment', 'nlp'],
    classifiers=['Development Status :: 4 - Beta',
                 'License :: OSI Approved :: MIT License',
                 'Environment :: Console',
                 'Operating System :: Microsoft :: Windows',
                 'Operating System :: POSIX :: Linux',
                 'Topic :: Text Processing',
                 'Programming Language :: Python :: 3.6'],
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    install_requires=REQUIREMENTS,
)



#!/usr/bin/env python
# -*-coding:utf-8-*-


from . import __softname__

DEFAULT_DICT = 'dict_small.txt'

DEFAULT_HMM_DATA = 'hmm_data.json'

DEFALUT_CACHE_NAME = f'{__softname__}.cache'

CACHE_WRITING = {}


#!/usr/bin/env python
# -*-coding:utf-8-*-

import time
import re
import os
from copy import deepcopy
import logging
import threading
from math import log
from nltk.tokenize.api import TokenizerI

from .base import BaseSegment
from fenci.train_hmm import train_emit_matrix, train_trans_matrix
from fenci.utils import strdecode, get_json_value, set_json_value
from fenci.const import CACHE_WRITING, DEFAULT_HMM_DATA
from fenci import __softname__

logger = logging.getLogger(__name__)

start_P = {'B': -0.26268660809250016,
           'E': -3.14e+100,
           'M': -3.14e+100,
           'S': -1.4652633398537678}

re_han_hmm = re.compile("([\u4E00-\u9FD5]+)")
re_skip_hmm = re.compile("([a-zA-Z0-9]+(?:\.\d+)?%?)")

MIN_FLOAT = -3.14e100

PrevStatus = {
    'B': 'ES',
    'M': 'MB',
    'S': 'SE',
    'E': 'BM'
}


class HMMSegment(TokenizerI, BaseSegment):
    def __init__(self, traning_root=None,
                 traning_regexp='.*\.txt', traning_mode='update',
                 cache_file=None):
        self.lock = threading.RLock()

        self.training_root = traning_root
        self.training_regexp = traning_regexp

        self.training_mode = traning_mode

        assert self.training_mode in ['update', 'replace']

        self.cache_file = cache_file
        self.tmp_dir = None

        self.P_trans = None
        self.model_data = {}
        self.P_emit = None

        self.initialized = False

    def __cut(self, sentence):
        self.check_initialized()

        prob, pos_list = viterbi(sentence, 'BMES', start_P, self.P_trans,
                                 self.P_emit)
        begin, nexti = 0, 0
        # logger.debug pos_list, sentence
        for i, char in enumerate(sentence):
            pos = pos_list[i]
            if pos == 'B':
                begin = i
            elif pos == 'E':
                yield sentence[begin:i + 1]
                nexti = i + 1
            elif pos == 'S':
                yield char
                nexti = i + 1
        if nexti < len(sentence):
            yield sentence[nexti:]

    def cut(self, sentence):
        sentence = strdecode(sentence)
        blocks = re_han_hmm.split(sentence)
        for blk in blocks:
            if re_han_hmm.match(blk):
                for word in self.__cut(blk):
                    yield word

            else:
                tmp = re_skip_hmm.split(blk)
                for x in tmp:
                    if x:
                        yield x

    def lcut(self, s):
        return list(self.cut(s))

    def tokenize(self, s):
        return self.lcut(s)

    def save_model(self):
        wlock = CACHE_WRITING.get('hmm_data', threading.RLock())
        CACHE_WRITING['hmm_data'] = wlock
        cache_file = self._get_cache_file()

        with wlock:
            logger.debug(
                "Dumping HMM model to file cache {0}".format(cache_file))
            set_json_value(cache_file, 'P_emit', self.model_data.get('P_emit'))
            set_json_value(cache_file, 'P_trans',
                           self.model_data.get('P_trans'))
            set_json_value(cache_file, 'hmm_timestamp', int(time.time()))

        try:
            del CACHE_WRITING['hmm_data']
        except KeyError:
            pass

    def training(self, root=None, regexp=None, training_mode='update'):
        assert training_mode in ['update', 'replace']

        if root is None and self.training_root is None:
            raise Exception('please give the training data root')
        root = root if root is not None else self.training_root
        regexp = regexp if regexp is not None else self.training_regexp
        training_mode = training_mode if training_mode is not None else self.training_mode

        if training_mode == 'update':
            P_emit = train_emit_matrix(root, regexp)
            P_trans = train_trans_matrix(root, regexp)
            old_P_trans = self.model_data.get('P_trans')
            old_P_emit = self.model_data.get('P_emit')

            new_P_emit = self.merge_P_emit(P_emit, old_P_emit)
            new_P_trans = self.merge_P_trans(P_trans, old_P_trans)
            self.model_data = {'P_emit': new_P_emit, 'P_trans': new_P_trans}
            self.P_emit = self._prepare_P_emit()
            self.P_trans = self._prepare_P_trans()

        elif training_mode == 'replace':
            P_emit = train_emit_matrix(root, regexp)
            P_trans = train_trans_matrix(root, regexp)
            self.model_data = {'P_emit': P_emit, 'P_trans': P_trans}
            self.P_emit = self._prepare_P_emit()
            self.P_trans = self._prepare_P_trans()

    def merge_P_trans(self, one, two):
        P_transMatrix = {'B': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                         'E': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                         'M': {'B': 0, 'E': 0, 'M': 0, 'S': 0},
                         'S': {'B': 0, 'E': 0, 'M': 0, 'S': 0}}

        from itertools import product
        for key in map(lambda a: a[0] + a[1],
                       product(['B', 'M', 'E', 'S'], repeat=2)):
            a = key[0]
            b = key[1]
            if a in one and b in one[a]:
                P_transMatrix[a][b] += one[a][b]
            if a in two and b in two[a]:
                P_transMatrix[a][b] += two[a][b]

        new_P_transMatrix = {}

        for k in P_transMatrix:
            for k2 in P_transMatrix[k]:
                if P_transMatrix[k][k2] == 0:
                    pass
                else:
                    if k not in new_P_transMatrix:
                        new_P_transMatrix[k] = {}
                    new_P_transMatrix[k][k2] = P_transMatrix[k][k2]
        return new_P_transMatrix

    def merge_P_emit(self, one, two):
        P_emit = {'B': {}, 'E': {}, 'M': {}, 'S': {}}

        for k, v in one.items():
            for word in v:
                P_emit[k][word] = v[word]

        for k, v in two.items():
            for word in v:
                if word in P_emit:
                    P_emit[k][word] += v[word]
                else:
                    P_emit[k][word] = v[word]

        return P_emit

    def initialize(self):
        with self.lock:
            if self.initialized:
                return

            t1 = time.time()

            cache_file = self._get_cache_file()

            # use cache data
            use_cache_data = False
            if os.path.isfile(cache_file):
                hmm_timestamp = get_json_value(cache_file,
                                               'hmm_timestamp')
                if hmm_timestamp:
                    use_cache_data = True

            if use_cache_data:
                logger.debug(
                    "Loading HMM model from cache {0}".format(cache_file))
                P_trans = get_json_value(cache_file, 'P_trans')
                P_emit = get_json_value(cache_file, 'P_emit')
                self.model_data = {'P_emit': P_emit, 'P_trans': P_trans}

                self.P_emit = self._prepare_P_emit()
                self.P_trans = self._prepare_P_trans()
            else:
                P_trans = get_json_value(self._get_default_model_file(),
                                         'P_trans')
                P_emit = get_json_value(self._get_default_model_file(),
                                        'P_emit')
                self.model_data = {'P_emit': P_emit, 'P_trans': P_trans}

                self.P_emit = self._prepare_P_emit()
                self.P_trans = self._prepare_P_trans()

                self.save_model()

            self.initialized = True
            logger.debug(
                "Loading model cost %.3f seconds." % (time.time() - t1))
            logger.debug("Prefix dict has been built succesfully.")

    def _get_default_model_file(self):
        from pkg_resources import resource_filename
        return resource_filename(__softname__, DEFAULT_HMM_DATA)

    def _prepare_P_trans(self):
        P_trans_data = self.model_data.get('P_trans')

        P_trans = deepcopy(P_trans_data)

        for k, v in P_trans.items():
            count = sum(v.values())
            for k2 in v:
                P_trans[k][k2] = log(P_trans[k][k2] / count)

        return P_trans

    def _prepare_P_emit(self):
        P_emit_data = self.model_data.get('P_emit')

        P_emit = deepcopy(P_emit_data)

        for k, v in P_emit.items():
            count = sum(v.values())
            for k2 in v:
                P_emit[k][k2] = log(P_emit[k][k2] / count)

        return P_emit


def viterbi(obs, states, start_p, trans_p, emit_p):
    V = [{}]  # tabular
    path = {}
    for y in states:  # init
        V[0][y] = start_p[y] + emit_p[y].get(obs[0], MIN_FLOAT)
        path[y] = [y]
    for t in range(1, len(obs)):
        V.append({})
        newpath = {}
        for y in states:
            em_p = emit_p[y].get(obs[t], MIN_FLOAT)
            (prob, state) = max(
                [(V[t - 1][y0] + trans_p[y0].get(y, MIN_FLOAT) + em_p, y0) for
                 y0 in PrevStatus[y]])
            V[t][y] = prob
            newpath[y] = path[state] + [y]
        path = newpath

    (prob, state) = max((V[len(obs) - 1][y], y) for y in 'ES')

    return (prob, path[state])
